"xvalid" <-
  function (geodata, coords = geodata$coords, data = geodata$data, 
            model, reestimate = FALSE, variog.obj = NULL,
            output.reestimate = FALSE, locations.xvalid = "all",
            messages.screen = TRUE, ...) 
{
  n <- nrow(coords)
  data <- as.vector(data)
  if (length(data) != n) 
    stop("coords and data have incompatible dimentions")
  xmat <- trend.spatial(trend = model$trend, coords = coords)
  ##
  ## Locations to be used in the cross-validation
  ##
  if(locations.xvalid == "all") locations.xvalid <- 1:n
  else{
    if(is.matrix(locations.xvalid) | is.data.frame(locations.xvalid))
      locations.xvalid <- is.vector(locations.xvalid)
    if(!is.vector(locations.xvalid) | !is.numeric(locations.xvalid))
      stop("argument locations.xvalid must be a numeric vector with numbers indicating the locations to be cross-validated")
    if(any(locations.xvalid) > n | length(locations.xvalid) > n)
      stop("incorrect value to the argument locations.xvalid.\nThis must be a numeric vector with numbers indicating the locations to be cross-validated")
  }
  n.pt.xv <- length(locations.xvalid)
  if(messages.screen){
    cat(paste("xvalid: number of data locations                 =", n))
    cat("\n")
    cat(paste("xvalid: number of locations for cross-validation =", n.pt.xv))
    cat("\n")
    cat("xvalid: performing cross-validation at location ... ")
    }
  ##
  ## Defining a function to predict at one point
  ## 
  cv.f <- function(ndata, ...) {
    if(messages.screen) cat(paste(ndata, ", ", sep=""))
    ## excluding data point
    coords.out <- coords[ndata, , drop = FALSE]
    data.out <- data[ndata]
    xmat.out <- xmat[ndata, , drop = FALSE]
    cv.coords <- coords[-ndata, ]
    cv.data <- as.vector(data)[-ndata]
    cv.xmat <- xmat[-ndata, , drop = FALSE]
    ## re-estimating the model
    if (reestimate) {
      if(model$method == "ML" | model$method == "REML" | model$method == "RML"){
        fix.pars <- (model$parameters.summary[c("tausq", "kappa", "psiA",
                                                "psiR", "lambda"), 1] == "fixed")
        val.pars <- model$parameters.summary[c("tausq", "kappa", 
                                               "psiA", "psiR", "lambda"), 2]
        names(fix.pars) <- c("tausq", "kappa", "psiA", "psiR", 
                             "lambda")
        names(val.pars) <- c("tausq", "kappa", "psiA", "psiR", 
                             "lambda")
        CVmod <- likfit(coords = cv.coords, data = cv.data, 
                        ini = model$cov.pars, fix.nugget = fix.pars["tausq"], 
                        nugget = val.pars["tausq"], fix.kappa = fix.pars["kappa"], 
                        kappa = val.pars["kappa"], fix.psiR = fix.pars["psiR"], 
                        psiR = val.pars["psiR"], fix.psiA = fix.pars["psiA"], 
                        psiA = val.pars["psiA"], fix.lambda = fix.pars["lambda"], 
                        lambda = val.pars["lambda"], cov.model = model$cov.model, 
                        trend = ~cv.xmat + 0, method = model$method, 
                        messages.screen = F, ...)
        if(output.reestimate){
          CVpars <- (CVmod$parameters.summary[c("tausq", "kappa", "psiA", "psiR", "lambda"), 2])
          CVpars <- c(CVmod$cov.pars, CVpars[fix.pars == FALSE])
        } 
      }
      if(model$method == "OLS" | model$method == "WLS"){
        if(is.null(variog.obj))
          stop("an object with the estimated variogram must be provided in the argument variog.obj when argument reestimate = TRUE")
        CVvar <- variog(coords = cv.coords, data = cv.data, uvec = variog.obj$uvec,
                        trend = variog.obj$trend, lambda = variog.obj$lambda,
                        option = variog.obj$output.type,
                        estimator.type = variog.obj$estimator.type,
                        nugget.tolerance = variog.obj$nugget.tolerance,
                        max.dist = max(variog.obj$u), pairs.min = 2,
                        bin.cloud = FALSE, direction = variog.obj$direction,
                        tolerance = variog.obj$tolerance,
                        unit.angle = "radians",
                        messages.screen = FALSE, ...)
        if(is.null(model$call$lower)) model$call$lower <- 0
        if(model$method == "OLS")
          CVmod <- olsfit(vario = CVvar, ini=model$cov.pars, cov.model = model$cov.model,
                          fix.nugget = model$fix.nugget, nugget = model$nugget,
                          kappa = model$kappa, max.dist = model$max.dist,
                          minimisation.function = model$minimisation.function,
                          lower = model$call$lower, messages.screen = FALSE, ...)
        if(model$method == "WLS"){
          if(is.null(model$call$weight)) model$call$weight <- "npairs"
          CVmod <- wlsfit(vario = CVvar, ini=model$cov.pars, cov.model = model$cov.model,
                          fix.nugget = model$fix.nugget, nugget = model$nugget,
                          kappa = model$kappa, max.dist = model$max.dist,
                          minimisation.function = model$minimisation.function,
                          lower = model$call$lower, weight = model$call$weight,
                          messages.screen = FALSE, ...)
        }
        if(output.reestimate){
          CVpars <- CVmod$cov.pars
          if(CVmod$fix.nugget == FALSE) CVpars <- c(CVpars, CVmod$nugget)
        }
      }
    }
    else CVmod <- model
    if(model$method == "ML" | model$method == "REML" | model$method == "RML"){
      fix.pars <- (CVmod$parameters.summary[c("tausq", "kappa", 
                                              "psiA", "psiR", "lambda"), 1] == "fixed")
      val.pars <- CVmod$parameters.summary[c("tausq", "kappa", 
                                             "psiA", "psiR", "lambda"), 2]
    }
    if(model$method == "OLS" | model$method == "WLS"){
      fix.pars <- c(CVmod$fix.nugget, T,T,T,T)
      if(is.null(CVmod$kappa)) CVmod$kappa <- 0.5
      val.pars <- c(CVmod$nugget, CVmod$kappa, 0, 1, CVmod$lambda)
    }
    names(fix.pars) <- c("tausq", "kappa", "psiA", "psiR", "lambda")
    names(val.pars) <- c("tausq", "kappa", "psiA", "psiR", "lambda")
    kr <- krige.conv(coords = cv.coords, data = cv.data, loc = coords.out,
                     krige = krige.control(trend.d = ~cv.xmat + 0,
                       trend.l = ~xmat.out + 0, cov.model = CVmod$cov.model, 
                       cov.pars = CVmod$cov.pars, nugget = CVmod$nugget, 
                       kappa = val.pars["kappa"], lambda = val.pars["lambda"], 
                       aniso.pars = val.pars[c("psiA", "psiR")]), mess = FALSE)
    res <- c(data.out, kr$pred, kr$krige.var)
    if(output.reestimate) res <- c(res, CVpars)
    ##, err = (data.out - kr$pred), e.rel = (data.out - kr$pred)/sqrt(kr$krige.var), 
    ##pval = pnorm(data.out, mean = kr$pred, sd = sqrt(kr$krige.var)))
    return(res)
  }
  res <- as.data.frame(t(apply(matrix(locations.xvalid), 1, cv.f)))
  if(messages.screen) cat("\nxvalid: end of cross-validation\n")
  if(output.reestimate){
    pars.names <- c("sigmasq", "phi")
    if(model$method == "ML" | model$method == "REML" | model$method == "RML"){
      fix.pars <- (model$parameters.summary[c("tausq", "kappa", 
                                              "psiA", "psiR", "lambda"), 1] == "fixed")
      pars.names <- c(pars.names,(c("tausq", "kappa", "psiA", "psiR", "lambda"))[fix.pars == FALSE])
    }
    if(model$method == "OLS" | model$method == "WLS")
      if(model$fix.nugget == FALSE) pars.names <- c(pars.names, "tausq")
    names(res) <- c(c("data", "pred", "krige.var"), pars.names)
  }
  else names(res) <- c("data", "predicted", "krige.var")
  res$error <- res$data - res$pred
  res$std.error <- res$err/sqrt(res$krige.var)
  res$prob <- pnorm(res$data, mean = res$pred, sd = sqrt(res$krige.var))
  if(output.reestimate){
    np <- length(pars.names)
    res <- res[,c((1:3), ((3+np+1):(6+np)),(4:(3+np)))] 
  }
  return(res)
}
